/*
 * WatchAndReason.hpp
 *
 *  Created on: Aug 24, 2009
 *      Author: kottlers
 */

#ifndef ASSIGNINFOS_HPP_
#define ASSIGNINFOS_HPP_
#define assigninfos_hpp_version      "$Revision: 1.9 $"
#include "BitVector.hpp"
#include "Types.hpp"
#include "Settings.hpp"
#include "BitVector.hpp"
#include "ThreadGlobals.hpp"
#include "Cls.hpp"


class VarVal{
	TwoBitsVec<Variable> vals;
public:
	VarVal():vals(){}
	VarVal(const VarVal& vls,const bool deep):vals(vls.vals,deep){}
	uint32_t 		count() const{return vals.getSize();}
	/*bool			takeThisIf(const VarVal& v){
					if(v.vals.getSize() == vals.getSize()){ return false;}
					vals = v.vals; return true;}*/
	void			ensureVarVals(uint32_t sz){
					asrt(sz > vals.getSize());
					vals.ensureSize(sz,true);}

	/*void			initWith(const VarVal& from){
					vals = from.vals;}*/


	bool			varHasValue(const Variable& v, bool b)const{
					return vals.hasVal(v,2+(uint32_t)b);}

	bool			varPrevValWasTrue(const Variable& v)const {
					return vals.secSet(v);}

	void			litSetPrevVal(const Literal& l){
					l.isPos() ? vals.setSec(l.toVar()) : vals.clearSec(l.toVar());}

	bool			litPrevValWas(const Literal& l)const {
					return vals.secSet(l.toVar()) == l.isPos();}

	void			varClearValue(const Variable& v){
					vals.clearFirst(v);}

	void			varClearValueFlipPrev(const Variable& v){
					vals.setVal(v,!vals.secSet(v));}

	bool			varKnwn(const Variable& v)const{
					return vals.firstSet(v);}

		//Literals
	bool			litTrue(const Literal& l)const 	{return varHasValue(l.toVar(),l.isPos());}
	bool			litFalse(const Literal& l)const	{return varHasValue(l.toVar(),l.isNeg());}
	bool			litUkwn(const Literal& l)const	{return !varKnwn(l.toVar());}
	bool			litKnwn(const Literal& l)const	{return varKnwn(l.toVar());}

	void			setValue(const Variable& var, const bool& pos){
					vals				.setVal(var,2+((uint32_t)pos));
					asrt(varKnwn(var));	}

	bool			setValueChng(const Variable& var, const bool& pos){
				uint64_t prev = vals.setGetVal(var,2+((uint32_t)pos));asrt(varKnwn(var)); asrt(prev < 2);
				return prev != pos;}

	void			setValue(const Literal& lit){
					setValue(lit.toVar(), lit.isPos());}

};


class BDReason{
protected:
	static const uint32_t DEC 		= 0;
	static const uint32_t CLS 		= 1;
	static const uint32_t BIN 		= 2;

	//state the kind of reason
	uint64_t	type 	: 	2; //Decision / Cls / Binary

	uint64_t	reason	: 	31; //bin or clp

	uint64_t	lbd		:	4;

	//decision level
	uint64_t	level	:	27;




public:
	static const uint32_t LBD_MAX 	= 16;
	BDReason():type(DEC),reason(max32>>1),lbd(LBD_MAX-1),level(max32>>5){}
	bool		isBinReason()		const	{ return type == BIN;}
	bool		isShortReason()		const	{ return type & 2;}
	bool		isClsReason()		const	{ return type == CLS;}
	bool		isDecision()		const	{ return type == DEC;}
	//bool		getHookVal()		const	{ return hook;}
	bool		isReason(const ClsP cls)const
				{return isClsReason() && cls== Ptr32(reason);}

	Literal		getBinReason()		const	{ asrt(isShortReason());return Literal(reason);}
	ClsP		getClsReason()  	const	{ asrt(isClsReason());return ClsP(reason);}
	uint32_t	getLevel()			const	{return level;}
	uint32_t	getLBD()			const	{return lbd+1;}
	//Literal	getDominator()		const	{return dom;}

	void 		setBinReason(const uint32_t lev, const Literal& l){asrt(level!=0);
				level=lev; type = BIN; reason= l.idx(); lbd=LBD_MAX-1; }
	void 		setClsReason(const uint32_t lev, const ClsP cls, const uint32_t LBD){asrt(level!=0); asrt(LBD>0); asrt(LBD<=LBD_MAX);
				level=lev; type = CLS; reason = cls.idx(); lbd=LBD-1;}
	void 		setDecision(const uint32_t lev){asrt(level!=0);
				level=lev; type = DEC; lbd=LBD_MAX-1;}
	void		setLevel(const uint32_t lev){asrt(level!=0);
				level=lev;	}
	void		makeFix(){level =0; type=DEC;}
	bool		isFix()const {return level==0;}
	//void		setHookVal(uint32_t b){asrt(b<2); hook=b;}
	//void		setDominator(const Literal& l){dom=l;}

	//void		setLevel(uint32_t m){asrt(level != 0); level=m;}
};



//template<class Reason>
class AssReason{
	AssReason(const AssReason& v){asrt(false);}
	AssReason& operator=(const AssReason& v){asrt(false);return *this;}
protected:
	einfo<BDReason,Variable>	reasons;
public:
	AssReason(){};
	uint32_t varsCount() const {return reasons.size();}

	void	ensureVars(uint32_t sz){
			asrt(sz > reasons.size());
			reasons			.ensureSize(sz);}


	uint32_t varAssLev(const Variable& v) const {
			return reasons[v].getLevel();}

	void	varSetAssLevl(const Variable& v, const uint32_t& l){
			reasons[v].setLevel(l); asrt(reasons[v].getLevel() == l);}

	bool 	varFix(const Variable& v) const {
			return reasons[v].isFix();}

	bool	varIsDecision(const Variable& v) const {
			return reasons[v].isDecision();}

	bool	varHasReason(const Variable& v,const ClsP cls) const{
			return reasons[v].isReason(cls);}

	bool	varIsShortReason(const Variable& v) const{
			return reasons[v].isShortReason();}

	const BDReason& varAssignReason(const Variable& v) const{
			return reasons[v];}

	/*const Literal varDominator(const Variable& v) const{
			return reasons[v].getDominator();}*/

	Literal	varGetShortFrstReason(const Variable& v) const{
			asrt(varIsShortReason(v));
			return reasons[v].getBinReason();}


	/*Literal	varGetShortSecReason(const Variable& v) const{
			asrt(varIsShortReason(v));
			return reasons[v].getShortSecondReason();}*/

	void	makeCurrFix(const Variable& var){
			reasons[var]		.makeFix();}

	//Literals
	//bool				varHook(const Variable& v)const 		{return reasons[v].getHookVal();}
	//bool				litHooked(const Literal& l)const		{return varHook(l.toVar()) == l.isPos();}
	//void				setHooked(const Literal& l)				{reasons[l.toVar()].setHookVal(l.isPos());}
	//void				setHookFromPrev(const Variable& v)		{reasons[v].setHookVal(varPrevValWasTrue(v));}
	const uint32_t		litAssLev(const Literal& l)		const 	{return varAssLev(l.toVar());}

	bool 				litFix(const Literal& l) 		const {return varFix(l.toVar());}
};







//template<class Reason>
class Assigns : public AssReason{
	VarVal&		vals;
	var_hp_t& 	act_hp;
	uint32_t		agility;
	const bool 	bump_agil;

	void setVal(const Variable& v, const bool p){
		agility -= agility/10000;
		if(vals.setValueChng(v,p)){
			agility += 1000;
			if(bump_agil){act_hp.activityBump(v);}
		}
	}

public:
	bool lessAgilThan(const uint32_t& x) const {return agility < x;}
	Assigns(VarVal& vls,var_hp_t& v_act_hp,const bool bump)
		:AssReason(),vals(vls),act_hp(v_act_hp),agility(0),bump_agil(bump){};
	VarVal& getValsRef() {return vals;}
	uint32_t reasonLBD(const Variable& var) const{
			asrt(reasons.inRange(var)); asrt(vals.varKnwn(var));
			return reasons[var].getLBD();
	}

	void	setValueDecision(const Literal& lit, uint32_t lev){
			const Variable var 	= lit.toVar(); asrt(reasons.inRange(var));
			const Literal lf	= lit.flipped();
			vals.setValue(var,lit.isPos());
								asrt(vals.varKnwn(lit.toVar()));asrt(vals.varPrevValWasTrue(lit.toVar())==lit.isPos());
			reasons[var]		.setDecision(lev);
			//roots[lf].lit		= lf.idx();
	}


	void	setValueReason(const Literal& lit, uint32_t lev,const ClsP cls, const uint32_t lbd){
			const Variable var 	= lit.toVar(); asrt(reasons.inRange(var));
			const Literal lf	= lit.flipped();
			vals.setValue(var,lit.isPos());
								asrt(vals.varKnwn(lit.toVar()));asrt(vals.varPrevValWasTrue(lit.toVar())==lit.isPos());
			reasons[var]		.setClsReason(lev,cls,lbd);
			//roots[lf].lit		= lf.idx();
			}

	void	setValueReason(const Literal& lit, uint32_t lev,const Literal& r1){//, const Literal& root){
			const Variable var 	= lit.toVar(); asrt(reasons.inRange(var));
			vals.setValue(var,lit.isPos());
								asrt(vals.varKnwn(lit.toVar()));asrt(vals.varPrevValWasTrue(lit.toVar())==lit.isPos());
			reasons[var]		.setBinReason(lev,r1);
			//roots[lit.flipped()].lit = root.idx();
			}

};



/*
bool BinVarInfo::asrtAllFlagsClear(){
	for(uint32_t i=roots.size();i--;){
		if(roots.byIndex(i).flag){
			std::cerr << "Literal "<< Literal(i).toDIMACS() << " is still flagged" << std::endl;
			return false;}
	}
	return true;
}
*/

//typedef VarInfo<BDReason> BinVarInfo;
//typedef VarInfo<BTReason> BinTernVarInfo;







#endif /* WATCHANDREASON_HPP_ */
