/*
 * Watching.hpp
 *
 *  Created on: 02.11.2010
 *      Author: stephan
 */

#ifndef WATCHING_HPP_
#define WATCHING_HPP_
#define watching_hpp_version "$Revision: 1.16 $"

#include "VecOfLists.hpp"
#include "AssignInfos.hpp"
#include <fstream>
#include "ThreadGlobals.hpp"
#include "Cls.hpp"



struct WtchEle{
	uint32_t wtchd;
	WtchEle(const Literal& l, const bool bin):wtchd((l.idx() << 1)| bin){}
	Literal getLit() const	{return Literal(wtchd>>1);}
	bool isBinary() const	{return wtchd & 1;}

	WtchEle(const ClsP& clp) : wtchd(clp.idx()){}
	ClsP getClsP() const	{return ClsP(wtchd);}
};

inline bool operator==(const WtchEle& a, const WtchEle& b){return a.wtchd==b.wtchd;}


class Watching{
	const VarVal&						vals;
	TGlobals&							glob;
	VecOfLists<WtchEle,Literal>			watches;

public:
	Watching(const VarVal& vls,TGlobals& g):vals(vls),glob(g),watches('W'){}
	inline bool bcpFindNewWatcher(const ClsP& ptr, const Literal& false_lit, const Literal& other);
	inline void clsCreateBinary(const Literal lit0, const Literal lit1);
	inline void clsCreateBinPart(const Literal& a, const Literal& b);
	inline void clsAddWtched(const Literal& to, const Literal& w, const ClsP& ptr);
	inline uint32_t writeDimacs(std::ofstream& out, bool print_lrnt);

	template<class Visitor, class LitFlag>
	inline void BFS(const Literal& start, Visitor& visi, LitFlag& flags, const bool visit_start, const bool clear_flags,vector<Literal>& queue);
	template<class LitFlag>
	void clearBfsFlags(LitFlag& flags, vector<Literal>& queue);
	template<class CollVisi,class MarkVisi, class LitFlag>
	void collectCommon(Literal lits[], uint32_t lsz,CollVisi& coll, MarkVisi& mark, LitFlag& flags, vector<Literal>& common_out);



	//void keepBin(WIter<WtchEle>& iter, const WtchEle& bin);
	//void keepCls(WIter<WtchEle>& iter, const WtchEle& oth, const WtchEle& cls);

	WIter<WtchEle> getWtchOfW(const Literal& l)	{return watches.getWIter(l);}
	Iter<WtchEle> getWtchOf(const Literal& l)		{return watches.getIter(l);}
	Literal 		getLitWithSmallerList(const Variable& v)	{return watches.sizeOf(v.toPosLit()) < watches.sizeOf(v.toNegLit()) ? v.toPosLit() : v.toNegLit();}
	bool isEmpty(const Literal& lit) const		{return watches.sizeOf(lit)==0;}
	void finish(const Literal& lit, WIter<WtchEle> it){watches.finish(lit,it);}
	void keepRest(const Literal& lit, WIter<WtchEle>& iter){
			while(iter.hasNext()){iter.keepNext();}
			watches.finish(lit,iter);}
	void clearOf(const Literal& l){watches.clear(l);}

	void enlarge(const uint32_t& sz){watches.enlarge(sz);}
	void reserveForTotal(const uint32_t& sz){ watches.reserveForTotal(sz);}
	uint32_t litsCount() const {return watches.listCount();}
	uint32_t getTotMem()const{return watches.getMemTotal();}
	void hardClear(){watches.hardClear();}

	inline bool asrtWtchingOK() const;
	inline bool asrtFindWtched(const Literal& lit, const ClsP& cand) const;
	inline bool asrtFindWtched(const Literal& lit, const Literal& cand) const;
};



bool	Watching::bcpFindNewWatcher(const ClsP& ptr, const Literal& false_lit, const Literal& other){

	Cls& cl 				= glob.getCls(ptr);
	/*uint32_t k 			= cl.last;
	const uint32_t stop 	= k;
	const uint32_t uwc	= cl.unWatchdCnt(); asrt(stop < uwc);
	do{
		uint32_t m = (++k == uwc);
		k &= (m-1);
		if(!vals.litFalse(cl[k])){
			cl		.setWatchers(other,cl[k],k); LOG_MEDIUM("Found new watcher "<< cl[k].toDIMACS());
			cl[k] 	= false_lit;
			clsAddWtched(cl.otherWatcher(other),other,ptr);
			return true;
		}
	}while(k!=stop);*/

	/*const Literal*  end = cl.uwEnd();
	for(Literal* lit = cl.uwBeg();lit != end;++lit){
		if(!vals.litFalse(*lit)){
			cl		.setWatchers(other,*lit); LOG_MEDIUM("Found new watcher "<< lit->toDIMACS());
			*lit 	= false_lit;
			clsAddWtched(cl.otherWatcher(other),other,ptr);
			return true;
		}
	}*/
	ClsLit* clit = cl.uwBeg();
	do{
		if(!vals.litFalse(Literal(clit->lit))){
			cl		.changeWatchers(other,clit,false_lit); LOG2("Found new watcher "<<cl.otherWatcher(other).toDIMACS());
			clsAddWtched(cl.otherWatcher(other),other,ptr);
			return true;
		}
	}while((clit++)->has_next);

	return false;
}

void Watching::clsCreateBinPart(const Literal& a, const Literal& b){
	watches.add(a,WtchEle(b,true));
}

void Watching::clsCreateBinary(const Literal lit0, const Literal lit1){

	watches.add(lit0,WtchEle(lit1,true));
	watches.add(lit1,WtchEle(lit0,true));
   	//++nbr_bin_clses;
	cnt_stat(bin_lts,2);
}

void	Watching::clsAddWtched(const Literal& to, const Literal& w, const ClsP& ptr){
	watches.add(to,WtchEle(w,false));
	watches.add(to,WtchEle(ptr));
}



uint32_t Watching::writeDimacs(std::ofstream& out, bool print_lrnt){

	uint32_t clses = 0;
	WtchEle we(INVALID_LIT,true); Literal oth; bool binary;
	for(uint32_t i=0;i!=watches.listCount();++i){
		Literal lit(i);
		Iter<WtchEle> iter = watches.getIter(lit);
		while(iter.getNext(we)){
			binary = we.isBinary();
			if(binary){
				oth = we.getLit();
			}else{
				asrt(iter.hasNext());
				iter.getNext(we);
				Cls& cls 	= glob.getCls(we.getClsP());
				if(!print_lrnt && cls.isLrnt()) continue;
				oth = cls.otherWatcher(lit);
			}

			if(oth < lit) continue;

			out << lit.toDIMACS() << " " << oth.toDIMACS() << " ";
			if(!binary){
				ClsLit* clit = glob.getCls(we.getClsP()).uwBeg();
				do{ out << clit->getLit().toDIMACS() << " ";}while((clit++)->has_next);
			}
			out << "0" << endl;
			++clses;
		}
	}
	return clses;
}




//for each target the Visitor functor is called exactly once
//the second argument is the predecessor by which target was reached first -> hence not all edges are called with visitor
//if visit_start is true the start literal is visited with predecessor "FALSE_LIT"

template<class Visitor, class LitFlag>
void Watching::BFS(const Literal& start, Visitor& visi, LitFlag& flags, const bool visit_start, const bool clear_flags,vector<Literal>& queue){
	asrt(!glob.elim.isSet(start.toVar()));	asrt(!clear_flags || queue.empty());
	LOG2("BFS run with start " << start.toDIMACS());
	uint32_t qhead = queue.size(); WtchEle we(INVALID_LIT,true);
	if(!visit_start || visi(start,FALSE_LIT).isTrue()){
		flags.setFlag(start); queue.push_back(start);}

	Literal target;
	while(qhead != queue.size()){
		Literal src 		= queue[qhead++]; //asrt(!dag.isLitRepl(src));
		Iter<WtchEle>ii	= getWtchOf(src.flipped());

		while(ii.getNext(we)){
			if(!we.isBinary()){asrt(ii.hasNext());ii.skipNext(); continue;}
			target = we.getLit(); asrt(glob.elim.isSet(target.toVar()));

			if(flags.isSet(target)){		 continue;}
			TrBool v_res				= visi(target,src);
			if(v_res.isUnknown()){		qhead=queue.size();break;}
			if(v_res.isTrue()){
				flags.setFlag(target);	queue.push_back(target);}
		}
	}

	LOG2("Finishing BFS run with start " << start.toDIMACS());
	if(clear_flags){clearBFSFlags(flags,queue);}
}

template<class LitFlag>
void Watching::clearBfsFlags(LitFlag& flags, vector<Literal>& queue){
	for(uint32_t i=queue.size();i--;){flags.clearFlag(queue[i]);}
	queue.clear();
}




template<class CollVisi,class MarkVisi, class LitFlag>
void Watching::collectCommon(Literal lits[], uint32_t lsz, CollVisi& coll, MarkVisi& mark, LitFlag& flags, vector<Literal>& common_out){

	if(lsz<2){	if(lsz==1){common_out.push_back(lits[0]);} return;}
	LOG2("Start search for common literals of " << lsz);
	vector<Literal> queue;

	BFS(lits[--lsz],coll,flags,true,true,queue);
	if(coll.common.empty()){LOG2("No common lits ");return;}

	while(lsz--){	asrt(!coll.common.empty());
		BFS(lits[lsz],mark,flags,true,false,queue);
		uint32_t cp = 0;
		for(uint32_t i=0;i!=coll.common.size();++i){
			if(flags.isSet(coll.common[i].first)){
				coll.common[cp++]=coll.common[i];LOG2("Keep lit "<<coll.common[i].first.toDIMACS());}
		}
		clearBFSFlags(flags,queue);
		if(cp == 0){LOG2("No common lits ");return;}
		coll.common.resize(cp);
	}
	asrt(!coll.common.empty());
	LOG2("By now common lits: " << coll.common.size());

	//take only topmost common literals -> others are redundant
	for(uint32_t i=coll.common.size();i--;){flags.setFlag(coll.common[i].first);LOG2("Flag lit"<<coll.common[i].first.toDIMACS());}
	for(uint32_t i=coll.common.size();i--;){
		if(!flags.isSet(coll.common[i].second)){common_out.push_back(coll.common[i].first);}
		else{LOG2("Dont take common " << coll.common[i].first.toDIMACS() << " because predecessor is taken: " << coll.common[i].second.toDIMACS());}
		flags.clearFlag(coll.common[i].first);
	}

	LOG2("Found common literals: " << common_out.size());
	asrt(!common_out.empty());
}


bool Watching::asrtFindWtched(const Literal& lit, const ClsP& cand) const{
	Iter<WtchEle> it = watches.getIter(lit);
	WtchEle we(FALSE_LIT,false);
	while(it.getNext(we)){
		if(we.isBinary()){continue;}
		if(!it.getNext(we)){ LOG0("Bad watching number in " << lit.toDIMACS()); return false;}
		if(we.getClsP()==cand) return true;
	}
	return false;
}


bool Watching::asrtFindWtched(const Literal& lit, const Literal& cand) const{
	Iter<WtchEle> it = watches.getIter(lit);
	WtchEle we(FALSE_LIT,false);
	while(it.getNext(we)){
		if(!we.isBinary()){
			if(!it.hasNext()){LOG0("Bad watching number in lit " << lit.toDIMACS()); return false;}
			it.skipNext();
			continue;
		}
		if(we.getLit()==cand) return true;
	}
	return false;
}

bool Watching::asrtWtchingOK() const{
	bool ok = true; WtchEle we(FALSE_LIT,false);
	Literal ol; ClsP oc;

	for(uint32_t i=watches.listCount();i--;){
		Literal lit(i);
		Iter<WtchEle> it = watches.getIter(lit);
		while(it.getNext(we)){
			if(we.isBinary()){
				ol = we.getLit();
				if(!asrtFindWtched(ol,lit)){ ok = false;
					LOG0("Binary " << lit.toDIMACS() << " and " << ol.toDIMACS() << " only one direction");}
			}else{
				if(!it.getNext(we)){LOG0("Bad watching number in lit " << lit.toDIMACS()); return false;}
				ClsP clp = we.getClsP();
				Cls& cls = glob.getCls(clp);
				ol		 = cls.otherWatcher(lit);
				if(!asrtFindWtched(ol,clp)){ok = false;
					LOG0("Clause " << clp.idx() << " only linked by " << lit.toDIMACS() << " but not by " << ol.toDIMACS());}
			}
		}
	}
	LOG1("Watching::asrtWtchingOK is " << (ok?"ok":"bad"));
	return ok;
}


#endif /* WATCHING_HPP_ */
